@@ -4,7 +4,7 @@ import os
 import sys
 from abc import ABC, abstractmethod
 from contextlib import contextmanager
-from os.path import expandvars, isabs
+from os.path import isabs
 
 from ...config.constants import AppEnvVars
 from ...utils.structures import EnvVars
@@ -52,6 +52,8 @@ class EnvironmentInterface(ABC):
         self.verbosity = verbosity
         self.__app = app
 
+        self._context = None
+
         self._system_python = None
         self._env_vars = None
         self._env_include = None
@@ -126,6 +128,15 @@ class EnvironmentInterface(ABC):
         """
         return self.__config
 
+    @property
+    def context(self):
+        if self._context is None:
+            from ...utils.context import Context
+
+            self._context = Context(self.root)
+
+        return self._context
+
     @property
     def system_python(self):
         if self._system_python is None:
@@ -647,22 +658,27 @@ class EnvironmentInterface(ABC):
         """
         Spawn a [shell](../config/hatch.md#shell) within the environment.
 
-        The shell should reflect any
-        [environment variables](environment.md#hatch.env.plugin.interface.EnvironmentInterface.get_env_vars)
-        the user defined either currently or at the time of
-        [creation](environment.md#hatch.env.plugin.interface.EnvironmentInterface.create).
+        This should either use
+        [command_context](environment.md#hatch.env.plugin.interface.EnvironmentInterface.command_context)
+        directly or provide the same guarantee.
         """
-        with self.get_env_vars():
+        with self.command_context():
             self.platform.exit_with_command([path])
 
-    def run_shell_commands(self, commands: list[str]):
+    def run_shell_command(self, command: str):
         """
-        This should yield the standard library's
+        This should return the standard library's
         [subprocess.CompletedProcess](https://docs.python.org/3/library/subprocess.html#subprocess.CompletedProcess)
-        for each command. Additionally, the commands must first be
-        [resolved](environment.md#hatch.env.plugin.interface.EnvironmentInterface.resolve_commands).
+        and will always be called when the
+        [command_context](environment.md#hatch.env.plugin.interface.EnvironmentInterface.command_context)
+        is active, with the expectation of providing the same guarantee.
+        """
+        return self.platform.run_command(command, shell=True)
 
-        The command execution should reflect any
+    @contextmanager
+    def command_context(self):
+        """
+        A context manager that when active should make executed shell commands reflect any
         [environment variables](environment.md#hatch.env.plugin.interface.EnvironmentInterface.get_env_vars)
         the user defined either currently or at the time of
         [creation](environment.md#hatch.env.plugin.interface.EnvironmentInterface.create).
@@ -670,42 +686,28 @@ class EnvironmentInterface(ABC):
         For an example, open the default implementation below:
         """
         with self.get_env_vars():
-            for command in self.resolve_commands(commands):
-                yield self.platform.run_command(command, shell=True)
+            yield
 
     def resolve_commands(self, commands: list[str]):
         """
         This expands each command into one or more commands based on any
         [scripts](../config/environment.md#scripts) that the user defined.
-        Each expanded command is then
-        [finalized](environment.md#hatch.env.plugin.interface.EnvironmentInterface.finalize_command).
         """
         for command in commands:
-            expanded_commands = self.expand_command(command)
-
-            for expanded_command in expanded_commands:
-                yield self.finalize_command(expanded_command)
-
-    def finalize_command(self, command: str):
-        """
-        Called for every
-        [resolved command](environment.md#hatch.env.plugin.interface.EnvironmentInterface.resolve_commands)
-        with the default behavior being equivalent to the standard library's
-        [os.path.expandvars](https://docs.python.org/3/library/os.path.html#os.path.expandvars).
-        """
-        return expandvars(command)
+            yield from self.expand_command(command)
 
     def expand_command(self, command):
         possible_script, _, remaining = command.partition(' ')
 
+        # Indicate undefined
+        if not remaining:
+            remaining = None
+
         if possible_script in self.scripts:
-            if remaining:
-                for cmd in self.scripts[possible_script]:
-                    yield f'{cmd} {remaining}'
-            else:
-                yield from self.scripts[possible_script]
+            for cmd in self.scripts[possible_script]:
+                yield self.context.format(cmd, args=remaining).strip()
         else:
-            yield command
+            yield self.context.format(command, args=remaining).strip()
 
     def construct_build_command(
         self,
